from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import random
import numpy as np
import tensorflow as tf

import grammar_check
import nltk
from nltk.corpus import brown


import readfile
import pronouncing
import string

flags = tf.flags
logging = tf.logging

flags.DEFINE_string(
    "model", "small",
    "A type of model. Possible options are: large.")
flags.DEFINE_string("data_path", None,
                    "Where the training/test data is stored.")
flags.DEFINE_string("save_path", None,
                    "Model output directory.")
flags.DEFINE_bool("use_fp16", False,
                  "Train using 16-bit floats instead of 32bit floats")

flags.DEFINE_string("get_pred", False,
                    "Only get predictions")

FLAGS = flags.FLAGS

tool = grammar_check.LanguageTool('en-GB')

class Node(object):

    def __init__(self, score, syllables, parent, word, words, state, preds, sentence):
        self.children = []
        self.score = score
        self.syllables = syllables
        self.parent = parent
        self.word = word
        self.words =words
        self.state = state
        self.preds = preds
        self.sentence = sentence

    def add_child(self, node):
        assert isinstance(node, Node)
        self.children.append(node)

    def remove_children(self):
      self.children = []

    def get_number_of_children(self):
      return len(self.children)

    def get_children(self):
      return self.children

    def __str__(self, level=0):
        ret = "\t"*level+repr(self.words)+"\n"
        for child in self.children:
            ret += child.__str__(level+1)
        return ret

    def get_word(self):
      return self.word





def data_type():
  return tf.float16 if FLAGS.use_fp16 else tf.float32


class LyrInput(object):

  def __init__(self, config, data, name=None):
    self.batch_size = batch_size = config.batch_size
    self.num_steps = num_steps = config.num_steps
    self.epoch_size = ((len(data) // batch_size) - 1) // num_steps
    self.input_data, self.targets = readfile.lyr_producer(data, batch_size, num_steps, name=name)


class LyrModel(object):

  def __init__(self, is_training, config, input_):
    self._input = input_

    batch_size = input_.batch_size
    num_steps = input_.num_steps
    size = config.hidden_size
    vocab_size = config.vocab_size

    self.x = input_.input_data

    def lstm_cell():
      return tf.contrib.rnn.BasicLSTMCell(size, forget_bias=0.0, state_is_tuple=True)

    attn_cell = lstm_cell

    if is_training and config.keep_prob < 1:
      def attn_cell():
        return tf.contrib.rnn.DropoutWrapper(lstm_cell(), output_keep_prob=config.keep_prob)

    cell = tf.contrib.rnn.MultiRNNCell([attn_cell() for _ in range(config.num_layers)], state_is_tuple=True)

    self._initial_state = cell.zero_state(batch_size, data_type())

    with tf.device("/cpu:0"):
      embedding = tf.get_variable("embedding", [vocab_size, size], dtype=data_type())
      inputs = tf.nn.embedding_lookup(embedding, self.x)

    if is_training and config.keep_prob < 1:
      inputs = tf.nn.dropout(inputs, config.keep_prob)

    inputs = tf.unstack(inputs, num=num_steps, axis=1)
    outputs, state = tf.contrib.rnn.static_rnn(cell, inputs, initial_state=self._initial_state)

    output = tf.reshape(tf.concat(outputs, 1), [-1, size])
    #print(output)
    softmax_w = tf.get_variable("softmax_w", [size, vocab_size], dtype=data_type())
    softmax_b = tf.get_variable("softmax_b", [vocab_size], dtype=data_type())

    logits = tf.matmul(output, softmax_w) + softmax_b
    #print(logits)
    loss = tf.contrib.legacy_seq2seq.sequence_loss_by_example([logits],[tf.reshape(input_.targets, [-1])],[tf.ones([batch_size * num_steps], dtype=data_type())])
    self._cost = cost = tf.reduce_sum(loss) / batch_size
    self._final_state = state
    #(logits)

    self.predictions = tf.nn.softmax(logits)


    if not is_training:
      return

    self._lr = tf.Variable(0.0, trainable=False)
    tvars = tf.trainable_variables()
    grads, _ = tf.clip_by_global_norm(tf.gradients(cost, tvars), config.max_grad_norm)
    optimizer = tf.train.GradientDescentOptimizer(self._lr)
    self._train_op = optimizer.apply_gradients(zip(grads, tvars), global_step=tf.contrib.framework.get_or_create_global_step())

    self._new_lr = tf.placeholder(tf.float32, shape=[], name="new_learning_rate")

    self._lr_update = tf.assign(self._lr, self._new_lr)

  def assign_lr(self, session, lr_value):
    session.run(self._lr_update, feed_dict={self._new_lr: lr_value})

  @property
  def input(self):
    return self._input

  @property
  def initial_state(self):
    return self._initial_state

  @property
  def cost(self):
    return self._cost

  @property
  def final_state(self):
    return self._final_state

  @property
  def lr(self):
    return self._lr

  @property
  def train_op(self):
    return self._train_op


class Config(object):
  init_scale = 0.04
  learning_rate = 1.0
  max_grad_norm = 10
  num_layers = 2
  num_steps = 35
  hidden_size = 1100
  max_epoch = 14
  max_max_epoch = 53
  keep_prob = 0.35
  lr_decay = 1 / 1.15
  batch_size = 20
  vocab_size = 27601


def run_epoch(session, model, eval_op=None, verbose=False):
  start_time = time.time()
  costs = 0.0
  iters = 0
  state = session.run(model.initial_state)

  fetches = {
      "cost": model.cost,
      "final_state": model.final_state,
      "preds" :model.predictions
  }
  if eval_op is not None:
    fetches["eval_op"] = eval_op

  #print('hei')

  for step in range(model.input.epoch_size):
    feed_dict = {}
    for i, (c, h) in enumerate(model.initial_state):
      feed_dict[c] = state[i].c
      feed_dict[h] = state[i].h

    vals = session.run(fetches, feed_dict)
    cost = vals["cost"]
    state = vals["final_state"]
    preds = vals["preds"]

    #print(preds)

    #print(state)


    costs += cost
    iters += model.input.num_steps


    if verbose and step % (model.input.epoch_size // 10) == 10:
      print("%.3f perplexity: %.3f speed: %.0f wps" %
            (step * 1.0 / model.input.epoch_size, np.exp(costs / iters),
             iters * model.input.batch_size / (time.time() - start_time)))

  return np.exp(costs / iters)


def get_config():
  if FLAGS.model == "conf":
    return Config()
  else:
    raise ValueError("Invalid model: %s", FLAGS.model)

def get_pred(session, model, word):
  fetches = {
      "cost": model.cost,
      "final_state": model.final_state,
      "preds" :model.predictions
  }

  feed_dict = {model.x: [[word]]}

  vals = session.run(fetches, feed_dict)
  cost = vals["cost"]
  state = vals["final_state"]
  preds = vals["preds"]

  return preds



def main(_):
  if not FLAGS.data_path:
    raise ValueError("Must set --data_path to data directory")

  raw_data = readfile.lyr_raw_data(FLAGS.data_path)
  train_data, valid_data, test_data, _, word_to_id = raw_data

  config = get_config()
  eval_config = get_config()
  eval_config.batch_size = 1
  eval_config.num_steps = 1

  with tf.Graph().as_default():
    initializer = tf.random_uniform_initializer(-config.init_scale, config.init_scale)

    with tf.name_scope("Train"):
      train_input = LyrInput(config=config, data=train_data, name="TrainInput")
      with tf.variable_scope("Model", reuse=None, initializer=initializer):
        m = LyrModel(is_training=True, config=config, input_=train_input)
      tf.summary.scalar("Training Loss", m.cost)
      tf.summary.scalar("Learning Rate", m.lr)

    with tf.name_scope("Valid"):
      valid_input = LyrInput(config=config, data=valid_data, name="ValidInput")
      with tf.variable_scope("Model", reuse=True, initializer=initializer):
        mvalid = LyrModel(is_training=False, config=config, input_=valid_input)
      tf.summary.scalar("Validation Loss", mvalid.cost)

    with tf.name_scope("Test"):
      test_input = LyrInput(config=eval_config, data=test_data, name="TestInput")
      with tf.variable_scope("Model", reuse=True, initializer=initializer):
        mtest = LyrModel(is_training=False, config=eval_config, input_=test_input)

    sv = tf.train.Supervisor(logdir=FLAGS.save_path)
    with sv.managed_session() as session:
      if not FLAGS.get_pred:
        for i in range(config.max_max_epoch):
          lr_decay = config.lr_decay ** max(i + 1 - config.max_epoch, 0.0)
          m.assign_lr(session, config.learning_rate * lr_decay)

          print("Epoch: %d Learning rate: %.3f" % (i + 1, session.run(m.lr)))

          train_perplexity = run_epoch(session, m, eval_op=m.train_op, verbose=True)
          print("Epoch: %d Train Perplexity: %.3f" % (i + 1, train_perplexity))

          valid_perplexity = run_epoch(session, mvalid)
          print("Epoch: %d Valid Perplexity: %.3f" % (i + 1, valid_perplexity))

        test_perplexity = run_epoch(session, mtest)
        print("Test Perplexity: %.3f" % test_perplexity)

        if FLAGS.save_path:
          print("Saving model to %s." % FLAGS.save_path)
          sv.saver.save(session, FLAGS.save_path + 'themodel', global_step=sv.global_step)

      else:

        id_to_word = {v: k for k, v in word_to_id.iteritems()}

        word = 'eos'
        num_words = 7

        state = None
        current_word = word_to_id[word]
        words = [current_word]
        vocab_size = 27601  

        top_score = [{'score': 0, 'p_id': 0, 'score_next': 0, 'state': None, 'preds': np.zeros(vocab_size)}]

        syllables1 = '10101010'


        related_words = []
        rhymes = []

        target = open('/home/loller/Desktop/RNN/dataset/relateandrhyme.txt', 'r')
        flag = 0
        for line in target:
          if line.startswith('---------'):
            flag = flag + 1
            continue
          else:
            if (flag == 0):
              rel = line.split(',')
              if ' ' not in rel[0]:
                related_words.append(rel[0])
            elif(flag == 1):
              rhyme = line.split(',')
              if ' ' not in rhyme[0] and ' ' not in rhyme[1]:
                rhymes.append([rhyme[0], rhyme[1]])
            elif(flag == 2):  
              rhyme = line.split(',')   
              if ' ' not in rhyme[0] and ' ' not in rhyme[1]:
                rhymes.append([rhyme[0], rhyme[1]])


        print(rhymes)


        target.close()

        feed_dict = {mtest.x: [[current_word]]}
        preds, state = session.run([mtest.predictions, mtest.final_state], feed_dict)

        root = Node(score = 0, syllables = '', parent = None, word = 'eos', words = 'eos', state = state, preds = preds, sentence = '')
        tree_list = [root]

        number_of_lines = 4
        rhyme_word_list = []


        #add stuff to rhyme_list

        not_done = True

        for i in range(len(rhymes)):
          if(rhymes[i][0] not in rhyme_word_list and rhymes[i][1] not in rhyme_word_list):
            if(len(rhymes[i][0]) > 1 and len(rhymes[i][1]) > 1):
              print(rhymes[i][0])
              print(rhymes[i][1])
              freq1 = nltk.FreqDist(t for w, t in brown.tagged_words() if w.lower() == rhymes[i][0])
              freq2 = nltk.FreqDist(t for w, t in brown.tagged_words() if w.lower() == rhymes[i][1])

              if (len(freq1.most_common()) > 0) and (len(freq2.most_common()) > 0):
                  rhyme_word_list.append(rhymes[i][0])
                  rhyme_word_list.append(rhymes[i][1])

          if(len(rhyme_word_list) == number_of_lines):
            break


        #rhyme_word_list = ['bird', 'absurd', 'blue', 'overdue']

        print(rhyme_word_list)


        for line_number in range(number_of_lines):

          print('kjorer')

          final_word = rhyme_word_list[line_number]
          final_id = word_to_id[final_word]

          syllables = syllables1[:-len(pronouncing.stresses(pronouncing.phones_for_word(final_word)[0]))] #cut syllables


          while(tree_list):
            look_at = tree_list[-1]
            got_score = False



            while(not got_score):
              syl = look_at.syllables #eg "101"


              p = np.squeeze(look_at.preds)


              p[0] = 0 #remove eos


              #higher score on related words:

              this_sent = look_at.sentence
              words = this_sent.split(' ')
              is_rel = any((True for x in related_words if x in words))

              rel_score = []

              if not is_rel:
                for x in related_words:
                  if(x not in look_at.words):
                    rel = {'score': p[word_to_id[x]], 'word': x}
                    if (len(rel_score) < 2):
                      rel_score.append(rel)
                    elif(rel['score'] > rel_score[1]['score']):
                      rel_score[1] = rel

                  rel_score = sorted(rel_score, key=lambda k: k['score'], reverse = True)
              

              #lower score on used words:
              used_words = look_at.words.split(' ')
              #print(used_words)
              for x in used_words:
                p[word_to_id[x]] = p[word_to_id[x]]/1000 



              #print(look_at.word)
              #print(p[final_id])




              p[np.argsort(p)[:-num_words]] = 0

              scores = []
              indexes = []

              for x in p:
                if(x != 0):
                  scores.append(x)
                  indexes.append(id_to_word[list(p).index(x)])

              if(len(rel_score) > 1):
                scores.append(rel_score[0]['score'])
                indexes.append(rel_score[0]['word'])
                scores.append(rel_score[1]['score'])
                indexes.append(rel_score[1]['word'])

              #print(indexes)
              word = ''
              #prune on score

              tot_lengde = len(syllables)

              #print(tot_lengde)
              #print(len(look_at.syllables))
              #print(look_at.score)
              #print(top_score[-1]['score'])
              #print((len(look_at.syllables)/tot_lengde)*top_score[-1]['score']/2)
              if((look_at.score >= ((len(look_at.syllables)/tot_lengde)*(7*top_score[-1]['score']/10))) or len(look_at.syllables)/tot_lengde <= 0.4):
                for item in indexes:
                  #print(item)
                  phones_list = pronouncing.phones_for_word(item)
                  #print(phones_list)
                  if(len(phones_list) > 0):
                    if(len(syllables) >= len(syl + pronouncing.stresses(phones_list[0]).replace('2', '1'))): #if(syllables.startswith(syl + pronouncing.stresses(phones_list[0]).replace('2', '1'))): #exact mattch. Maybe just check length. Or percentage
                      #print('get in')
                      if (item not in [o.get_word() for o in look_at.get_children()]):
                        word = item
                        #print(word)
                        break;
                  else:
                    continue #vil hoppe over denne i for loopen

              #print(word)
              if(word == ''): #did I look at all my paths? #doesnt find any more paths. goes in here. ends
                #print(look_at.syllables)
                #print(syllables)
                if(len(look_at.syllables) == len(syllables)): #if(look_at.syllables == syllables): 
                  if(len(top_score) < 12): #set length of top_score

                    #in text here, must look from last eos!

                    #text = (look_at.words + ' ' + final_word).replace('eos ', '').replace(' i ', ' I ').replace('i\'m', 'I\'m').replace('i ', 'I ')
                    #print(text)
                    #time.sleep(0.1)
                    #matches = tool.check(text)
                    #correct = grammar_check.correct(text, matches)
                    #print(correct)

                    #if (text == correct):


                    #only do this if end_word_wc.most_common() er tom


                    end_word_wc = nltk.FreqDist(t for w, t in brown.tagged_words() if w.lower() == final_word)

                    sentence = look_at.words.replace('eos', '.') + ' ' + final_word

                    tokens = nltk.word_tokenize(sentence)

                    p_t = nltk.pos_tag(tokens)


                    pos_end = ''
                    for index in p_t:
                      if index[0] == final_word:
                        pos_end = index[1]

                    keep = False

                    for indexx in end_word_wc.most_common():
                      if (pos_end in indexx):
                        keep = True
                        break

                    if keep:
                      new_dict = {'score': look_at.score, 'p_id': look_at.words, 'state': look_at.state, 'preds': np.squeeze(look_at.preds), 'score_next': np.squeeze(look_at.preds)[final_id]}
                      top_score.append(new_dict)
                      top_score = sorted(top_score, key=lambda k: k['score'], reverse = True)

                  else:

                    #beregne score

                    sentence_score = look_at.score
                    sentence_next_score = np.squeeze(look_at.preds)[final_id]

                    sent_score_list = []
                    next_score_list = []

                    for x in top_score:
                      sent_score_list.append(x['score'])
                      next_score_list.append(x['score_next'])

                    sent_score_list.append(sentence_score)
                    next_score_list.append(sentence_next_score)

                    #print('ding')

                    #print(sentence_score)
                    #print(sentence_next_score)

                    #print('-----')

                    #print(top_score[0]['score'])
                    #print(top_score[1]['score'])
                    #print(top_score[2]['score'])

                    #print('-----')

                    #print(top_score[0]['score_next'])
                    #print(top_score[1]['score_next'])
                    #print(top_score[2]['score_next'])

                    #print('-----')

                    #print(sent_score_list)
                    #print(next_score_list)



                    #normalize

                    normed1 = [i/sum(sent_score_list) for i in sent_score_list]

                    if sum(next_score_list) != 0.0:
                      next_score_list = [i/sum(next_score_list) for i in next_score_list]

                    print('-----')
                    print(normed1)
                    print(next_score_list)

                    this_score = next_score_list.pop() + normed1.pop()

                    final_score_list = []

                    for i in range(len(normed1)):
                      final_score_list.append(normed1[i] + next_score_list[i])

                    min_val = min(final_score_list)

                    #print('-----')
                    #print(this_score)

                    #print(final_score_list)

                    #print(min_val)


                    if(this_score > min_val): #elif(look_at.score > top_score[2]['score']):
                      #text = (look_at.words + ' ' + final_word).replace('eos ', '').replace(' i ', ' I ').replace('i\'m', 'I\'m')
                      #matches = tool.check(text)
                      #correct = grammar_check.correct(text, matches)
                      #if (text == correct):

                      end_word_wc = nltk.FreqDist(t for w, t in brown.tagged_words() if w.lower() == final_word)

                      sentence = look_at.words.replace('eos', '.') + ' ' + final_word



                      tokens = nltk.word_tokenize(sentence)

                      p_t = nltk.pos_tag(tokens)


                      pos_end = ''
                      for index in p_t:
                        if index[0] == final_word:
                          pos_end = index[1]



                      keep = False

                      for indexx in end_word_wc.most_common():
                        if (pos_end in indexx):
                          keep = True
                          break




                      if keep:
                        new_dict = {'score': look_at.score, 'p_id': look_at.words, 'state': look_at.state, 'preds': np.squeeze(look_at.preds), 'score_next': np.squeeze(look_at.preds)[final_id]}
                        top_score[final_score_list.index(min_val)] = new_dict
                        top_score = sorted(top_score, key=lambda k: k['score'], reverse = True)



                    #print(top_score) 
                got_score = True
                del tree_list[-1]
                #for child in look_at.get_children():
                # child.parent = None
                # del child
                look_at.remove_children()

              elif (look_at.syllables != syllables): #syllables are not yet done

                #what word, score, get preds and state

                feed_this = word_to_id[word]

                feed_dict = {mtest.x: [[feed_this]], mtest._initial_state: look_at.state}
                preds, state = session.run([mtest.predictions, mtest.final_state], feed_dict)

                #give better for rel

                sscore = look_at.score + scores[indexes.index(word)]

                for i in range(len(rel_score)):
                  if ((rel_score[i]['word'] in (look_at.sentence + ' ' + word))):
                    sscore = sscore + 0.4
                    break

                new_node = Node(score = sscore, syllables = look_at.syllables + pronouncing.stresses(pronouncing.phones_for_word(word)[0]).replace('2', '1'), 
                  word = word, words = look_at.words + ' ' + word, parent = look_at, preds = preds, state = state, sentence = look_at.sentence + ' ' + word)
                look_at.add_child(new_node)
                tree_list.append(new_node)
                look_at = tree_list[-1]
              print(str(root))


          #expand on end word here (choose from top list and expand on end word and eos)


          feed_this1 = final_id

          #print(top_score[0]['state'])



          #velg top_score obj

          last_list = []

          for x in top_score:
            if not any(d['p_id'].startswith(x['p_id'][:int(5* len(x['p_id']) / 6)]  ) for d in last_list):
              last_list.append(x)





      

          for entry in last_list:
            print(entry['p_id'])
            print(entry['score'])

          obj = random.choice(last_list)







          feed_dict = {mtest.x: [[feed_this1]], mtest._initial_state: obj['state']} #topscore.state
          preds, nstate = session.run([mtest.predictions, mtest.final_state], feed_dict)

          feed_this = word_to_id['eos']

          feed_dict = {mtest.x: [[feed_this]], mtest._initial_state: nstate} #the prev state.
          preds, rstate = session.run([mtest.predictions, mtest.final_state], feed_dict)

          #normalize the score


          root = Node(score = 0, syllables = '', parent = None, word = 'eos', words = obj['p_id'] + ' ' + id_to_word[feed_this1] + ' eos', state = rstate, preds = preds, sentence = '')


          
          #update score
          #flush top_score
          #set new root

          print('-------------')

          for entry in top_score:
            print(entry['p_id'])
            print(entry['score'])

          print(root.words)
    
          top_score = [{'score': 0, 'p_id': 0, 'state': None, 'preds': np.zeros(vocab_size), 'score_next': 0}]
          tree_list = [root]




































            
        #print(id_to_word)

        #for i in range(num_words):


         # if state is not None:
          #  feed_dict = {mtest.x: [[current_word]], mtest._initial_state: state}
          #else:
          #  feed_dict = {mtest.x: [[current_word]]}

          #preds, state = session.run([mtest.predictions, mtest.final_state], feed_dict)

          #p = np.squeeze(preds)
          #p[np.argsort(p)[:-pick_top_words]] = 0
          #p = p / np.sum(p)
          #current_word = np.random.choice(vocab_size, 1, p=p)[0]

          #current_word = np.argmax(p)
          #this_score = max(p)

          #print(p)
          #total_score = total_score + this_score

          #words.append(current_word)

        #print(total_score)
        #words = map(lambda x: id_to_word[x], words)
        #print(" ".join(words).decode('utf-8'))







        







if __name__ == "__main__":
  tf.app.run()
